c*********************************************************************
c 
c File           qubic.f
c Author         Jelena Ivanisevic
c Date           October, 1996
c
c Description    Qubic Equation solver
c
c
c*********************************************************************
c
c   Qubic Equation solver
c
c    p(x) = a(3)*x**3 + a(2)*x**2 + a(1)*x + a(0)
c  
c    a        is a real array of dimension 0..3 sucah that a(i) 
c             holds the coefficient i of the cubic p(x) above
c             The procedure qubic doesn't change a(i).
c
c    rr, ri   are both real arrays of the dimension 1..3 that 
c             hold the real and imaginary part of the i-th root
c             rr(i) real part of the root i
c             ri(i) imaginery part of the root i
c
c    iflag    integer returning one of the following values:
c
c              1  if the root in rr(i), ri(i) meets the specification
c
c              0  if the qubic has infinitely many roots
c
c             -1  if the i-th root doesn't exist because the true 
c                 degree of the polynomial is less then i
c  
c             -2  if any of the coefficaients are not normalized 
c             
c             -3  if the true i-th root is outside the range of norm.
c                 floating point numbers(including zero); in this 
c                 case rr(i) ri(i) may be +/=Inf a denormalized 
c                 number or zero. ( true root is not zero)
c
c             -4  if the true i-th root is within the range of 
c                 normalized numbers ( including 0 ) but at least one 
c                 of rr(i) ri(i) can not be guaranteed to meet the 
c                 specification  in this case rr(i) ri(i) which may 
c                 be any IEEE number except NaN contain rough approx
c                 imation to root i.
c
c             -5  can not guarantee that any of the iflag values above 
c                 is correct but root calculated is a rough approxima-
c                 tion to the root i which is any floating point number 
c                 except NaN
c
c             -6  can not guarantee that any of the iflag values above 
c                 is correct and root is not calculated. rr(i) ri(i)
c                 contain arbitary values.
c
c SPECIFICATION
c 
c Roots of the qubic shall be normalized floating point numbers
c that ar exact zeroes of the polynomial  
c 
c    p(x) = a'(3)*x**3 + a'(2)*x**2 + a'(1)*x + a'(0)
c
c where
c     a'(3) = a(3)
c     |a(0) - a'(0)| < K * meps * |a(0)|
c     |a(1) - a'(1)| < L * meps * max( |a(0)|/rmin, |a(1)| ) 
c     |a(2) - a'(2)| < M * meps * max( |a(3)|*rmax, |a(2)| )      
c
c and K, L, M, are small integers < 100
c rmin and rmax are bounds for the real root. One of the 
c ways to estimate them is:
c
c 
c     (|a(0)|/(a'+|a(0)|)) .le. |r| .le. (1+|a/a(3)|)
c
c     where 
c           a'= amaxh is max(|a(3)|, |a(2)|, |a(1)|)
c           a = amaxl  is max(|a(2)|, |a(1)|, |a(0)|)
c
c  
c Note 1:
c Derivation of these error relations will be presented in the 
c exernal documentation
c 
c Note 2:
c
c Degenerate cases will relay on quadretic solver specification.
c
c Note 3:
c  
c This note is valid if we use running error analysis suggested
c in [1] or equivalent and suggested composite deflation in [1]
c
c
c REFERENCES
c [1] Peters, Wilkinson ,"Practical Problems Arising ..." 
c [2] Numerical Recepies
c [3] Wilkinson "Rounding Errors in Algrebaic Processes"
c [4] Adams, "A Stopping Criteria ..."
c [5] Zaguskin "Handbook of Numerical Methods..."
c 
c IMPLEMENTATION
c 
c This subroutine solves qubic in the following manner:
c
c     1. Bracket real root ( qubic has at least one )
c     2. Use Newton-Raphson's method to find that root 
c     3. Use composed deflation to deflate qubic to quadratic
c     4. Solve quadratic
c
c Comments:
c The Newton Raphson has two 'regular' stop criteria:
c  1. root(k) = root(k-1)
c  2. Running roundoff error estimation
c 
c 
c Assumptions:
c Quadratic finds exact roots (specs 1.1a)  
c 
c Future improvements:
c 1. Bracketing can be improved, in [5] better critera can be found. 
c    Becouse of overflows I have restricted number of good roots
c    that can be calculated if this was more sistematically 
c    processed. 
c 2. May be Infinities and Nan's are not checked everywhere
c 3. Denormalized coefficients are not taken into account. I have 
c    realized too late that we are supposed to cover them as well.
c       
c JI, October 1996
c**********************************************************************             
             
      subroutine qubic( a, rr, ri, iflag )
 
      real rr(1:3), ri(1:3)
      integer iflag(1:3)
      real a(0:3)
      real s(0:2)
 
c
c     Set all flags to -6. 
c     
      iflag(1) = -6
      iflag(2) = -6
      iflag(3) = -6 
 
c
c     Check invalid input 
c     
 
      do i=0,3,1
         if ( isInf(a(i)) .or. isNaN(a(i)) .or. isDnrm(a(i))) then
            iflag(1)= -2
            iflag(2)= -2
            iflag(3)= -2
            return
         endif
      enddo
c
c Check zero coefficiens
c
      if ( a(3) .eq. 0. ) then 
 
c        Report that third root doesn't exist 
         iflag(3) = -1
         
c        Quadratic equation      
         call quad( a(2), a(1), a(0), rr(1), ri(1), rr(2), ri(2), nr)
c        Set iflag(2) iflag(1) based on nr
 
         if ( nr .eq. -2 ) then 
 
c           input coefficients are not normalized    
         
            iflag(1) = -2
            iflag(2) = -2
            
         else if ( nr .eq. -1 ) then 
 
c           there are infinitely many roots
 
            iflag(1) = 0
            iflag(2) = 0
            iflag(3) = 0
         
         else if ( nr .eq. 0 ) then 
c           there are no roots             
            iflag(1) = -1
            iflag(2) = -1
 
         else if ( nr .eq. 11 ) then 
 
c           11 if there is one root (i.e., a=0, b=/=0)
c           and it is a normalized number. 
c           (return the root in r1r, r1i with r2r, r2i arbitrary.)
c
            iflag(1) = 1
            iflag(2) = -1
 
         else if ( nr .eq. 10 )  then 
 
c           10 if there is one root (i.e., a=0, b=/=0) and it is not
c           a normalized number. 
c           (r1r, r1i contain an approximate root which might be 0, 
c           a denormalized number, Inf or NaN; r2r, r2i arbitrary.)
 
            iflag(1) = -3
            iflag(2) = -1
 
         else if ( nr .eq. 22 ) then 
 
c           22 if there are two roots (i.e., a=/=0) and both are
c           normalized numbers. 
c           (return one root in r1r, r1i and the other in r2r,r2i.)
 
            iflag(1) = 1
            iflag(2) = 1
 
         else if ( nr .eq. 21 ) then 
 
c           21 if there are two roots (i.e., a=/=0) and only one is
c           a normalized number. 
c           (return the one normalized root in r1r, r1i and an approx
c           root which might be 0, a denormalized number, Inf or NaN
c           in r2r, r2i.)
 
            iflag(1) = 1
            iflag(2) = -3
 
         else if ( nr .eq. 20 ) then
 
c           20 if there are two roots (i.e., a=/=0) and neither is
c           a normalized numbers.
c           (return approximations to the roots which might be 0, 
c           a denormalized number, Inf or NaN in r1r, r1i and r2r,r2i
c 
            iflag(1) = -3
            iflag(2) = -3
 
         endif
         return
      endif
 
c     Special case if a(0) = 0 , rr(1)= ri(1)= 0, iflag(1) = 0
      if ( a(0) .eq. 0. ) then 
c       Set first zero to 0. 
        rr(1) = 0.
        ri(1) = 0.
        iflag(1) = 1
 
c       Solve Quadratic 
        call quad( a(3), a(2), a(1), rr(2), ri(2), rr(3), ri(3), nr)
 
c       Set iflag(2) iflag(3) based on nr
 
         if ( nr .eq. 22 ) then 
 
c           22 if there are two roots (i.e., a=/=0) and both are
c           normalized numbers. 
c           (return one root in r1r, r1i and the other in r2r,r2i.)
 
            iflag(2) = 1
            iflag(3) = 1
 
         else if ( nr .eq. 21 ) then 
 
c           21 if there are two roots (i.e., a=/=0) and only one is
c           a normalized number. 
c           (return the one normalized root in r1r, r1i and an approx
c           root which might be 0, a denormalized number, Inf or NaN
c           in r2r, r2i.)
 
            iflag(2) = 1
            iflag(3) = -3
 
         else if ( nr .eq. 20 ) then 
 
c           20 if there are two roots (i.e., a=/=0) and neither is
c           a normalized numbers.
c           (return approximations to the roots which might be 0, 
c           a denormalized number, Inf or NaN in r1r, r1i and r2r,r2i
c 
            iflag(2) = -3
            iflag(3) = -3
 
         endif
        
        return   
      endif
 
 
c     TRUE QUBIC 
c     
c     a(3)=/0, excluding spec case a(0)=0. that has already been
c     solved. 
 
     
c
c       
c     Root bracketing
c
c     Root bracketing is a process of calculating interval
c     (a, b) such that f(a)*f(b)<0 i.e. function has opposite
c     signs in a and b. Generally, if we do not know the 
c     function this can be difficult task. 
c     The function that we are looking at is polynomial. 
c     Ther are few methods to determine the upper and lower 
c     bound of the poynimial roots. The simplest one is used in 
c     this routine. The weakness is that the bounds obtained
c     are not pretty weak. For better bounds see 
c     Suppose that we take   
c     x1 = 0.
c     x2 = +/- M where M is upper bound for roots
c     
c     Qubic has odd number of real zero crossings in this 
c     interval which means that f(0) and f(M) have to have 
c     opposite signs.
c     Therefore, this can be used as initial bracket for 
c     any zero-finding methods.  
c 
c     To calculate M use relation 
c
c     (|a(0)|/(a'+|a(0)|)) .le. |x| .le. (1+|a/a(3)|)
c
c     where 
c           a'= amaxh is max(|a(3)|, |a(2)|, |a(1)|)
c           a = amaxl  is max(|a(2)|, |a(1)|, |a(0)|)
c
      x1 = 0.
      amaxl = max(abs(a(2)), abs(a(1)), abs(a(0))) 
      x2 = 1. + amaxl/abs(a(3))
c
c     in x2 relation underflows are harmless. Overflows we 
c     can not handle so algorithm stops with an error. 
 
 
c     if both signs are same, there is a negative real root
c     if signs are different real root is positive.
    
      if ( sign(1., a(0)) * sign(1.,a(3)) .gt. 0.) then
c        look for the negative real root
         x2 = -x2 
      endif
c     
      if ( isInf( x2 ) ) then 
c        Bracketing failed
	 print *, 'bracketing'
         return 
      endif
 
c
c     Bracketing successful, proceed with NR zero finding  
c
 
c     NEWTON-RAPHSON 
 
c     Iterative root finding - Newton-Raphson Method
 
      r = rtsafe( x1, x2, iflagrt, a )

      print *, 'root=',r, 'flag=',iflagrt      

      rr(1) = r
      ri(1) = 0.
 
      if( iflagrt .ge. 0 ) then
 
c        Root found
 
         if ( isInf(r) ) then 
            iflag(1) = -3
            return 
         else if ( isNaN(r) ) then
	     print *, 'root NaN'
             return 
 
         else if ( isDnrm(r) .or. r .eq. 0. ) then 
             iflag(1) = -3
          else
            iflag(1) = 1
         endif
 
      else 
 
c        Root not found
 
c        RTsafe exceeding max number of iterations. '
c        Inf, Nans occured in calculations

	 print *, 'root not found' 
         return            
 
      endif
 
c
c     DEFLATION
     
      call deflation( r, a, s, iflagdf )
 
      if ( iflagdf .eq. -1 ) then
 
c        Due to Inf, Nans not able to deflate
         return 
      
      else if ( iflagdf .eq. 0 ) then 
 
c        Deflation not composite as in [1], do not guarantee that
c        specification will be met, adjust flags
      
         if ( isDnrm(r) .or. r .eq. 0. ) then 
             iflag(1) = -5
         else
             iflag(1) = -4
         endif    
      endif
 
c     SOLVING QUADRATIC
      
      call quad( s(2), s(1), s(0), rr(2), ri(2), rr(3), ri(3), nr )
 
c     Set flags iflag(2) iflag(3) according to the result of quadratic
 
c         if ( nr .eq. -2 ) then 
 
c           deflated coefficients are not normalized    
c           root 2, 3, not found
c            iflag(2) = -6
c            iflag(3) = -6
            
c         else if ( nr .eq. -1 ) then 
 
c        there are infinitely many roots
c        deflated coefficients all equal to 0
            
c            iflag(2) = -6
c            iflag(3) = -6
         
c         else if ( nr .eq. 0 )
 
c           there are no roots
c           in deflated polynomial
               
c            iflag(2) = -6
c            iflag(3) = -6
 
c         else
c
          
         if ( nr .eq. 11 ) then
c   
c           11 if there is one root (i.e., a=0, b=/=0)
c           and it is a normalized number. 
c           return the root in rr(2),ri(2) with rr(3),ri(3) arbitrary
c
            if ( iflag(1) .eq. 1. ) then 
c              first root was good root   
               iflag(2) = 1 
            else
c              first root was rough approx
               iflag(2) = -5 
            endif    
 
         else if ( nr .eq. 10 ) then
 
c           10 if there is one root (i.e., a=0, b=/=0) and it is not
c           a normalized number. 
c           (r1r, r1i contain an approximate root which might be 0, 
c           a denormalized number, Inf or NaN; r2r, r2i arbitrary.)
c
            if ( isNaN(rr(2)) .or. isNaN(ri(2)) ) then 
               iflag(2) = -6
            else if ( iflag(1) .eq. 1. ) then 
c              first root was good root   
               iflag(2) = -3 
            else
c              first root was rough approx
               iflag(2) = -5 
            endif    
            iflag(3) = -6
 
         else if ( nr .eq. 22 )  then
 
c           22 if there are two roots (i.e., a=/=0) and both are
c           normalized numbers. 
c           (return one root in rr,ri(2) and the other in rr,ri(3).)
c           iflags now depend on the accuracy of the first root. 
 
            if ( iflag(1) .eq. 1. ) then 
c              first root was good root   
               iflag(2) = 1 
               iflag(3) = 1
            else
c              first root was rough approx
               iflag(2) = -5 
               iflag(3) = -5
            endif    
              
            
         else if ( nr .eq. 21 )  then 
 
c           21 if there are two roots (i.e., a=/=0) and only one is
c           a normalized number. 
c           (return the one normalized root in r1r, r1i and an approx
c           root which might be 0, a denormalized number, Inf or NaN
c           in r2r, r2i.)
 
            if ( iflag(1) .eq. 1. ) then 
c              first root was good root   
               iflag(2) = 1 
               if ( isNaN(rr(3)) .or. isNaN(ri(3)) ) then 
                  iflag(3) = -3
               endif
            else
c              first root was rough approx
               iflag(2) = -5 
               if ( isNaN(rr(3)) .or. isNaN(ri(3)) ) then 
                  iflag(3) = -5
               endif
            endif    
 
         else if ( nr .eq. 20 )  then 
 
c           20 if there are two roots (i.e., a=/=0) and neither is
c           a normalized numbers.
c           (return approximations to the roots which might be 0, 
c           a denormalized number, Inf or NaN in rr(2),ri(2),
c           rr(3),ri(3)
c 
            if ( iflag(1) .eq. 1. ) then 
c              first root was good root   
               if ( isNaN(rr(2)) .or. isNaN(ri(2)) ) then 
                  iflag(2) = -3
               endif
               if ( isNaN(rr(3)) .or. isNaN(ri(3)) ) then 
                  iflag(3) = -3
               endif
            else
c              first root was rough approx
               if ( isNaN(rr(2)) .or. isNaN(ri(2)) ) then 
                  iflag(2) = -5
               endif 
               if ( isNaN(rr(3)) .or. isNaN(ri(3)) ) then 
                  iflag(3) = -5
               endif
            endif    
 
         endif
         
          
      return
      end
